// SPDX-FileCopyrightText: Copyright (c) 2008-2018, NVIDIA Corporation. All rights reserved.
// SPDX-License-Identifier: Apache-2.0

/*! \file thrust/system/cuda/memory.h
 *  \brief Managing memory associated with Thrust's CUDA system.
 */

#pragma once

#include <thrust/detail/config.h>

#if defined(_CCCL_IMPLICIT_SYSTEM_HEADER_GCC)
#  pragma GCC system_header
#elif defined(_CCCL_IMPLICIT_SYSTEM_HEADER_CLANG)
#  pragma clang system_header
#elif defined(_CCCL_IMPLICIT_SYSTEM_HEADER_MSVC)
#  pragma system_header
#endif // no system header
#include <thrust/detail/type_traits.h>
#include <thrust/memory.h>
#include <thrust/mr/allocator.h>
#include <thrust/system/cuda/memory_resource.h>

#include <cuda/std/__host_stdlib/ostream>

THRUST_NAMESPACE_BEGIN
namespace cuda_cub
{
/*! Allocates an area of memory available to Thrust's <tt>cuda</tt> system.
 *  \param n Number of bytes to allocate.
 *  \return A <tt>cuda::pointer<void></tt> pointing to the beginning of the newly
 *          allocated memory. A null <tt>cuda::pointer<void></tt> is returned if
 *          an error occurs.
 *  \note The <tt>cuda::pointer<void></tt> returned by this function must be
 *        deallocated with \p cuda::free.
 *  \see cuda::free
 *  \see std::malloc
 */
inline _CCCL_HOST_DEVICE pointer<void> malloc(std::size_t n);

/*! Allocates a typed area of memory available to Thrust's <tt>cuda</tt> system.
 *  \param n Number of elements to allocate.
 *  \return A <tt>cuda::pointer<T></tt> pointing to the beginning of the newly
 *          allocated elements. A null <tt>cuda::pointer<T></tt> is returned if
 *          an error occurs.
 *  \note The <tt>cuda::pointer<T></tt> returned by this function must be
 *        deallocated with \p cuda::free.
 *  \see cuda::free
 *  \see std::malloc
 */
template <typename T>
inline _CCCL_HOST_DEVICE pointer<T> malloc(std::size_t n);

/*! Deallocates an area of memory previously allocated by <tt>cuda::malloc</tt>.
 *  \param ptr A <tt>cuda::pointer<void></tt> pointing to the beginning of an area
 *         of memory previously allocated with <tt>cuda::malloc</tt>.
 *  \see cuda::malloc
 *  \see std::free
 */
inline _CCCL_HOST_DEVICE void free(pointer<void> ptr);

/*! \p cuda::allocator is the default allocator used by the \p cuda system's
 *  containers such as <tt>cuda::vector</tt> if no user-specified allocator is
 *  provided. \p cuda::allocator allocates (deallocates) storage with \p
 *  cuda::malloc (\p cuda::free).
 */
template <typename T>
using allocator = thrust::mr::stateless_resource_allocator<T, thrust::system::cuda::memory_resource>;

//! \p cuda::universal_allocator allocates managed memory that can be used by the \p cuda system and host systems.
template <typename T>
using universal_allocator =
  thrust::mr::stateless_resource_allocator<T, thrust::system::cuda::universal_memory_resource>;

//! \p cuda::universal_host_pinned_allocator allocates pinned host memory that can be used by the \p cuda system and
//! host systems.
template <typename T>
using universal_host_pinned_allocator =
  thrust::mr::stateless_resource_allocator<T, thrust::system::cuda::universal_host_pinned_memory_resource>;
} // namespace cuda_cub

namespace system::cuda
{
using thrust::cuda_cub::allocator;
using thrust::cuda_cub::free;
using thrust::cuda_cub::malloc;
using thrust::cuda_cub::universal_allocator;
using thrust::cuda_cub::universal_host_pinned_allocator;
} // namespace system::cuda

/*! \namespace thrust::cuda
 *  \brief \p thrust::cuda is a top-level alias for \p thrust::system::cuda.
 */
namespace cuda
{
using thrust::cuda_cub::allocator;
using thrust::cuda_cub::free;
using thrust::cuda_cub::malloc;
using thrust::cuda_cub::universal_allocator;
using thrust::cuda_cub::universal_host_pinned_allocator;
} // namespace cuda

THRUST_NAMESPACE_END

#include <thrust/system/cuda/detail/memory.inl>
